from __future__ import print_function
from rosco.toolbox.ofTools.fast_io.output_processing import output_processing
import rosco.toolbox

def FAST_IO_timeseries(fname):
    # interface to FAST_IO data load
    try:
        test = rosco.toolbox.__file__
    except:
        print('WARNING: rosco.toolbox required for wisdem.aeroelasticse.FAST_post.FAST_IO_timeseries')
    
    fast_out = output_processing()
    fast_data = fast_out.load_fast_out(fname, verbose=True)[0]
    return fast_data



import os
import shutil
import subprocess

class TurbsimReader(object):

    def read_input_file(self, input_file_name):
        inpf = open(input_file_name, 'r')

        # Runtime Options
        inpf.readline()
        inpf.readline()
        inpf.readline()
        self.Echo = inpf.readline().split()[0]
        self.RandSeed1 = inpf.readline().split()[0]
        self.RandSeed2 = inpf.readline().split()[0]
        self.WrBHHTP = inpf.readline().split()[0]
        self.WrFHHTP = inpf.readline().split()[0]
        self.WrADHH = inpf.readline().split()[0]
        self.WrADFF = inpf.readline().split()[0]
        self.WrBLFF = inpf.readline().split()[0]
        self.WrADTWR = inpf.readline().split()[0]
        self.WrFMTFF = inpf.readline().split()[0]
        self.WrACT = inpf.readline().split()[0]
        self.Clockwise = inpf.readline().split()[0]
        self.ScaleIEC = inpf.readline().split()[0]

        # Turbine/Model Specifications
        inpf.readline()
        inpf.readline()
        self.NumGrid_Z = inpf.readline().split()[0]
        self.NumGrid_Y = inpf.readline().split()[0]
        self.TimeStep = inpf.readline().split()[0]
        self.AnalysisTime = inpf.readline().split()[0]
        self.UsableTime = inpf.readline().split()[0]
        self.HubHt = inpf.readline().split()[0]
        self.GridHeight = inpf.readline().split()[0]
        self.GridWidth = inpf.readline().split()[0]
        self.VFlowAng = inpf.readline().split()[0]
        self.HFlowAng = inpf.readline().split()[0]

        # Meteorological Boundary Conditions 
        inpf.readline()
        inpf.readline()
        self.TurbModel = inpf.readline().split()[0]
        self.UserFile = inpf.readline().split()[0].replace("'","").replace('"','')
        self.IECstandard = inpf.readline().split()[0]
        self.IECturbc = inpf.readline().split()[0]
        self.IEC_WindType = inpf.readline().split()[0]
        self.ETMc = inpf.readline().split()[0]
        self.WindProfileType = inpf.readline().split()[0]
        self.ProfileFile = inpf.readline().split()[0].replace("'","").replace('"','')
        self.RefHt = inpf.readline().split()[0]
        self.URef = inpf.readline().split()[0]
        self.ZJetMax = inpf.readline().split()[0]
        self.PLExp = inpf.readline().split()[0]
        self.Z0 = inpf.readline().split()[0]


        # Meteorological Boundary Conditions 
        inpf.readline()
        inpf.readline()
        self.Latitude = inpf.readline().split()[0]
        self.RICH_NO = inpf.readline().split()[0]
        self.UStar = inpf.readline().split()[0]
        self.ZI = inpf.readline().split()[0]
        self.PC_UW = inpf.readline().split()[0]
        self.PC_UV = inpf.readline().split()[0]
        self.PC_VW = inpf.readline().split()[0]

        # Spatial Coherence Parameters
        inpf.readline()
        inpf.readline()
        self.SCMod1 = inpf.readline().split()[0]
        self.SCMod2 = inpf.readline().split()[0]
        self.SCMod3 = inpf.readline().split()[0]
        self.InCDec1 = inpf.readline().split()[0]
        self.InCDec2 = inpf.readline().split()[0]
        self.InCDec3 = inpf.readline().split()[0]
        self.CohExp = inpf.readline().split()[0]

        # Spatial Coherence Parameters
        inpf.readline()
        inpf.readline()
        self.CTEventPath = inpf.readline().split()[0]
        self.CTEventFile = inpf.readline().split()[0]
        self.Randomize = inpf.readline().split()[0]
        self.DistScl = inpf.readline().split()[0]
        self.CTLy = inpf.readline().split()[0]
        self.CTLz = inpf.readline().split()[0]
        self.CTStartTime = inpf.readline().split()[0]


class TurbsimWriter(object):

    def __init__(self, turbsiminputs):

        self.turbsiminputs = turbsiminputs
        
    def execute(self, turbsim_input_file):

         tsinp = open(turbsim_input_file, 'w')
         tsinp.write("---------TurbSim v2.00.* Input File------------------------\n")
         tsinp.write(" Turbsim input file\n")
         tsinp.write("---------Runtime Options-----------------------------------\n")

         # runtime options
         tsinp.write('{!s:<12}   Echo            - Echo input data to <RootName>.ech (flag)\n'.format(self.turbsiminputs.Echo))
         tsinp.write('{!s:<12}   RandSeed1       - First random seed  (-2147483648 to 2147483647)\n'.format(int(self.turbsiminputs.RandSeed1)))
         tsinp.write('{!s:<12}   RandSeed2       - Second random seed (-2147483648 to 2147483647) for intrinsic pRNG, or an alternative pRNG: "RanLux" or "RNSNLW"\n'.format(self.turbsiminputs.RandSeed2))
         tsinp.write('{!s:<12}   WrBHHTP         - Output hub-height turbulence parameters in binary form?  (Generates RootName.bin)\n'.format(self.turbsiminputs.WrBHHTP))
         tsinp.write('{!s:<12}   WrFHHTP         - Output hub-height turbulence parameters in formatted form?  (Generates RootName.dat)\n'.format(self.turbsiminputs.WrFHHTP))
         tsinp.write('{!s:<12}   WrADHH          - Output hub-height time-series data in AeroDyn form?  (Generates RootName.hh)\n'.format(self.turbsiminputs.WrADHH))
         tsinp.write('{!s:<12}   WrADFF          - Output full-field time-series data in TurbSim/AeroDyn form? (Generates RootName.bts)\n'.format(self.turbsiminputs.WrADFF))
         tsinp.write('{!s:<12}   WrBLFF          - Output full-field time-series data in BLADED/AeroDyn form?  (Generates RootName.wnd)\n'.format(self.turbsiminputs.WrBLFF))
         tsinp.write('{!s:<12}   WrADTWR         - Output tower time-series data? (Generates RootName.twr)\n'.format(self.turbsiminputs.WrADTWR))
         tsinp.write('{!s:<12}   WrFMTFF         - Output full-field time-series data in formatted (readable) form?  (Generates RootName.u, RootName.v, RootName.w)\n'.format(self.turbsiminputs.WrFMTFF))
         tsinp.write('{!s:<12}   WrACT           - Output coherent turbulence time steps in AeroDyn form? (Generates RootName.cts)\n'.format(self.turbsiminputs.WrACT))
         tsinp.write('{!s:<12}   Clockwise       - Clockwise rotation looking downwind? (used only for full-field binary files - not necessary for AeroDyn)\n'.format(self.turbsiminputs.Clockwise))
         tsinp.write('{!s:<12}   ScaleIEC        - Scale IEC turbulence models to exact target standard deviation? [0=no additional scaling; 1=use hub scale uniformly; 2=use individual scales]\n'.format(self.turbsiminputs.ScaleIEC))

         # Turbine/Model Specifications
         tsinp.write("\n")
         tsinp.write("--------Turbine/Model Specifications-----------------------\n")
         tsinp.write('{!s:<12}   NumGrid_Z       - Vertical grid-point matrix dimension\n'.format(self.turbsiminputs.NumGrid_Z))
         tsinp.write('{!s:<12}   NumGrid_Y       - Horizontal grid-point matrix dimension\n'.format(self.turbsiminputs.NumGrid_Y))
         tsinp.write('{!s:<12}   TimeStep        - Time step [seconds]\n'.format(self.turbsiminputs.TimeStep))
         tsinp.write('{!s:<12}   AnalysisTime    - Length of analysis time series [seconds] (program will add time if necessary: AnalysisTime = MAX(AnalysisTime, UsableTime+GridWidth/MeanHHWS) )\n'.format(self.turbsiminputs.AnalysisTime))
         tsinp.write('{!s:<12}   UsableTime      - Usable length of output time series [seconds] (program will add GridWidth/MeanHHWS seconds unless UsableTime is "ALL")\n'.format(self.turbsiminputs.UsableTime))
         tsinp.write('{!s:<12}   HubHt           - Hub height [m] (should be > 0.5*GridHeight)\n'.format(self.turbsiminputs.HubHt))
         tsinp.write('{!s:<12}   GridHeight      - Grid height [m]\n'.format(self.turbsiminputs.GridHeight))
         tsinp.write('{!s:<12}   GridWidth       - Grid width [m] (should be >= 2*(RotorRadius+ShaftLength))\n'.format(self.turbsiminputs.GridWidth))
         tsinp.write('{!s:<12}   VFlowAng        - Vertical mean flow (uptilt) angle [degrees]\n'.format(self.turbsiminputs.VFlowAng))
         tsinp.write('{!s:<12}   HFlowAng        - Horizontal mean flow (skew) angle [degrees]\n'.format(self.turbsiminputs.HFlowAng))

         # Meteorological Boundary Conditions
         tsinp.write("\n")
         tsinp.write("--------Meteorological Boundary Conditions-------------------\n")
         tsinp.write('{!s:<12}   TurbModel       - Turbulence model ("IECKAI","IECVKM","GP_LLJ","NWTCUP","SMOOTH","WF_UPW","WF_07D","WF_14D","TIDAL","API","USRINP","TIMESR", or "NONE")\n'.format(self.turbsiminputs.TurbModel))
         tsinp.write('{!s:<12}   UserFile        - Name of the file that contains inputs for user-defined spectra or time series inputs (used only for "USRINP" and "TIMESR" models)\n'.format(self.turbsiminputs.UserFile))
         tsinp.write('{!s:<12}   IECstandard     - Number of IEC 61400-x standard (x=1,2, or 3 with optional 61400-1 edition number (i.e. "1-Ed2") )\n'.format(self.turbsiminputs.IECstandard))
         tsinp.write('{!s:<12}   IECturbc        - IEC turbulence characteristic ("A", "B", "C" or the turbulence intensity in percent) ("KHTEST" option with NWTCUP model, not used for other models)\n'.format(self.turbsiminputs.IECturbc))
         tsinp.write('{!s:<12}   IEC_WindType    - IEC turbulence type ("NTM"=normal, "xETM"=extreme turbulence, "xEWM1"=extreme 1-year wind, "xEWM50"=extreme 50-year wind, where x=wind turbine class 1, 2, or 3)\n'.format(self.turbsiminputs.IEC_WindType))
         tsinp.write('{!s:<12}   ETMc            - IEC Extreme Turbulence Model "c" parameter [m/s]\n'.format(self.turbsiminputs.ETMc))
         tsinp.write('{!s:<12}   WindProfileType - Velocity profile type ("LOG";"PL"=power law;"JET";"H2L"=Log law for TIDAL model;"API";"USR";"TS";"IEC"=PL on rotor disk, LOG elsewhere; or "default")\n'.format(self.turbsiminputs.WindProfileType))
         tsinp.write('{!s:<12}   ProfileFile     - Name of the file that contains input profiles for WindProfileType="USR" and/or TurbModel="USRVKM" [-]\n'.format(self.turbsiminputs.ProfileFile))
         tsinp.write('{!s:<12}   RefHt           - Height of the reference velocity (URef) [m]\n'.format(self.turbsiminputs.RefHt))
         tsinp.write('{!s:<12}   URef            - Mean (total) velocity at the reference height [m/s] (or "default" for JET velocity profile) [must be 1-hr mean for API model; otherwise is the mean over AnalysisTime seconds]\n'.format(self.turbsiminputs.URef))
         tsinp.write('{!s:<12}   ZJetMax         - Jet height [m] (used only for JET velocity profile, valid 70-490 m)\n'.format(self.turbsiminputs.ZJetMax))
         tsinp.write('{!s:<12}   PLExp           - Power law exponent [-] (or "default")\n'.format(self.turbsiminputs.PLExp))
         tsinp.write('{!s:<12}   Z0              - Surface roughness length [m] (or "default")\n'.format(self.turbsiminputs.Z0))

         # Non-IEC Meteorological Boundary Conditions
         tsinp.write("\n")
         tsinp.write("--------Non-IEC Meteorological Boundary Conditions------------\n")
         tsinp.write('{!s:<12}   Latitude        - Site latitude [degrees] (or "default")\n'.format(self.turbsiminputs.Latitude))
         tsinp.write('{!s:<12}   RICH_NO         - Gradient Richardson number [-]\n'.format(self.turbsiminputs.RICH_NO))
         tsinp.write('{!s:<12}   UStar           - Friction or shear velocity [m/s] (or "default")\n'.format(self.turbsiminputs.UStar))
         tsinp.write('{!s:<12}   ZI              - Mixing layer depth [m] (or "default")\n'.format(self.turbsiminputs.ZI))
         tsinp.write('{!s:<12}   PC_UW           - Hub mean uw Reynolds stress [m^2/s^2] (or "default" or "none")\n'.format(self.turbsiminputs.PC_UW))
         tsinp.write('{!s:<12}   PC_UV           - Hub mean uv Reynolds stress [m^2/s^2] (or "default" or "none")\n'.format(self.turbsiminputs.PC_UV))
         tsinp.write('{!s:<12}   PC_VW           - Hub mean vw Reynolds stress [m^2/s^2] (or "default" or "none")\n'.format(self.turbsiminputs.PC_VW))

         # Spatial Coherence Parameters
         tsinp.write('\n')
         tsinp.write(
             '--------Spatial Coherence Parameters----------------------------\n')
         tsinp.write('{!s:<12}   SCMod1           - u-component coherence model ("GENERAL", "IEC", "API", "NONE", or "default")\n'.format(
             self.turbsiminputs.SCMod1))
         tsinp.write('{!s:<12}   SCMod2           - v-component coherence model ("GENERAL", "IEC", "NONE", or "default")\n'.format(
             self.turbsiminputs.SCMod2))
         tsinp.write('{!s:<12}   SCMod3           - w-component coherence model ("GENERAL", "IEC", "NONE", or "default")\n'.format(
             self.turbsiminputs.SCMod3))
         if not type(self.turbsiminputs.InCDec1) is str:
            tsinp.write('{:<5.2f}  {:<5.2f}   InCDec1        - u-component coherence parameters for general or IEC models [-, m^-1] (e.g. "10.0  0.3e-3" in quotes) (or "default")\n'.format(
                float(self.turbsiminputs.InCDec1[0]), float(self.turbsiminputs.InCDec1[1])))
         else:
            tsinp.write('{!s:<12}   InCDec1        - u-component coherence parameters for general or IEC models [-, m^-1] (e.g. "10.0  0.3e-3" in quotes) (or "default")\n'.format(
                self.turbsiminputs.InCDec1))
         if not type(self.turbsiminputs.InCDec2) is str:
            tsinp.write('{:<5.2f}  {:<5.2f}   InCDec2        - v-component coherence parameters for general or IEC models [-, m^-1] (e.g. "10.0  0.3e-3" in quotes) (or "default")\n'.format(
                float(self.turbsiminputs.InCDec2[0]), float(self.turbsiminputs.InCDec2[1])))
         else:
            tsinp.write('{!s:<12}   InCDec2        - v-component coherence parameters for general or IEC models [-, m^-1] (e.g. "10.0  0.3e-3" in quotes) (or "default")\n'.format(
                self.turbsiminputs.InCDec2))
         if not type(self.turbsiminputs.InCDec3) is str:
            tsinp.write('{:<5.2f}  {:<5.2f}   InCDec3        - w-component coherence parameters for general or IEC models [-, m^-1] (e.g. "10.0  0.3e-3" in quotes) (or "default")\n'.format(
                float(self.turbsiminputs.InCDec3[0]), float(self.turbsiminputs.InCDec3[1])))
         else:
            tsinp.write('{!s:<12}   InCDec3        - w-component coherence parameters for general or IEC models [-, m^-1] (e.g. "10.0  0.3e-3" in quotes) (or "default")\n'.format(
                self.turbsiminputs.InCDec3))
         tsinp.write('{!s:<12}   CohExp           - Coherence exponent for general model [-] (or "default")\n'.format(
             self.turbsiminputs.CohExp))

         # Coherent Turbulence Scaling Parameters
         tsinp.write('\n')
         tsinp.write('--------Coherent Turbulence Scaling Parameters-------------------\n')
         tsinp.write('{!s:<12}   CTEventPath     - Name of the path where event data files are located\n'.format(self.turbsiminputs.CTEventPath))
         tsinp.write('{!s:<12}   CTEventFile     - Type of event files ("LES", "DNS", or "RANDOM")\n'.format(self.turbsiminputs.CTEventFile))
         tsinp.write('{!s:<12}   Randomize       - Randomize the disturbance scale and locations? (true/false)\n'.format(self.turbsiminputs.Randomize))
         tsinp.write('{!s:<12}   DistScl         - Disturbance scale [-] (ratio of event dataset height to rotor disk). (Ignored when Randomize = true.)\n'.format(self.turbsiminputs.DistScl))
         tsinp.write('{!s:<12}   CTLy            - Fractional location of tower centerline from right [-] (looking downwind) to left side of the dataset. (Ignored when Randomize = true.)\n'.format(self.turbsiminputs.CTLy))
         tsinp.write('{!s:<12}   CTLz            - Fractional location of hub height from the bottom of the dataset. [-] (Ignored when Randomize = true.)\n'.format(self.turbsiminputs.CTLz))
         tsinp.write('{!s:<12}   CTStartTime     - Minimum start time for coherent structures in RootName.cts [seconds]\n'.format(self.turbsiminputs.CTStartTime))

         tsinp.close()



import pytest
import os.path as osp
import platform

# looking up OS for the correct executable extension
mactype = platform.system().lower()
if mactype in ["linux", "linux2", "darwin"]:
    exeExt = ""
elif mactype in ["win32", "windows", "cygwin"]: #NOTE: platform.system()='Windows', sys.platform='win32'
    libext = '.exe'
else:
    raise ValueError('Unknown platform type: '+mactype)

REPOSITORY_ROOT = osp.dirname(osp.dirname(osp.dirname(osp.dirname(__file__))))
BUILD_DIR = osp.join(REPOSITORY_ROOT, "build/reg_tests")

# Path to the OpenFAST executable
OF_PATH = osp.join(REPOSITORY_ROOT,"build/glue-codes/openfast",f"openfast{exeExt}")

def pytest_addoption(parser):
    parser.addoption("--executable", action="store", default=OF_PATH, help="Path to the OpenFAST executable")
    parser.addoption("--source_dir", action="store", default=REPOSITORY_ROOT, help="Path to the openfast repository")
    parser.addoption("--build_dir", action="store", default=BUILD_DIR, help="Path to the test data directory")




import pytest
import os.path as osp
import subprocess, sys


from openfast_io.FAST_reader import InputReader_OpenFAST
from openfast_io.FAST_writer import InputWriter_OpenFAST
from openfast_io.FAST_output_reader import FASTOutputFile

from openfast_io.FileTools import check_rtest_cloned, compare_fst_vt
from pathlib import Path

from conftest import REPOSITORY_ROOT, BUILD_DIR, OF_PATH



# Exercising the  various OpenFAST modules
FOLDERS_TO_RUN = [
    "AWT_YFix_WSt"                           , # "openfast;elastodyn;aerodyn;servodyn"
    "AWT_WSt_StartUp_HighSpShutDown"         , # "openfast;elastodyn;aerodyn;servodyn"
    "AWT_YFree_WSt"                          , # "openfast;elastodyn;aerodyn;servodyn"
    "AWT_YFree_WTurb"                        , # "openfast;elastodyn;aerodyn;servodyn"
    "AWT_WSt_StartUpShutDown"                , # "openfast;elastodyn;aerodyn;servodyn"
    "AOC_WSt"                                , # "openfast;elastodyn;aerodyn;servodyn"
    "AOC_YFree_WTurb"                        , # "openfast;elastodyn;aerodyn;servodyn"
    "AOC_YFix_WSt"                           , # "openfast;elastodyn;aerodyn;servodyn"
    "UAE_Dnwind_YRamp_WSt"                   , # "openfast;elastodyn;aerodyn;servodyn"
    "UAE_Upwind_Rigid_WRamp_PwrCurve"        , # "openfast;elastodyn;aerodyn;servodyn"
    "WP_VSP_WTurb_PitchFail"                 , # "openfast;elastodyn;aerodyn;servodyn"
    "WP_VSP_ECD"                             , # "openfast;elastodyn;aerodyn;servodyn"
    "WP_VSP_WTurb"                           , # "openfast;elastodyn;aerodyn;servodyn"
    "SWRT_YFree_VS_EDG01"                    , # "openfast;elastodyn;aerodyn;servodyn"
    "SWRT_YFree_VS_EDC01"                    , # "openfast;elastodyn;aerodyn;servodyn"
    "SWRT_YFree_VS_WTurb"                    , # "openfast;elastodyn;aerodyn;servodyn"
    "5MW_Land_DLL_WTurb"                     , # "openfast;elastodyn;aerodyn;servodyn"
    "5MW_Land_DLL_WTurb_wNacDrag"            , # "openfast;elastodyn;aerodyn;servodyn"
    "5MW_OC3Mnpl_DLL_WTurb_WavesIrr"         , # "openfast;elastodyn;aerodyn;servodyn;hydrodyn;subdyn;offshore"
    "5MW_OC3Mnpl_DLL_WTurb_WavesIrr_Restart" , # "openfast;elastodyn;aerodyn;servodyn;hydrodyn;subdyn;offshore;restart"
    "5MW_OC3Trpd_DLL_WSt_WavesReg"           , # "openfast;elastodyn;aerodyn;servodyn;hydrodyn;subdyn;offshore"
    "5MW_OC4Jckt_DLL_WTurb_WavesIrr_MGrowth" , # "openfast;elastodyn;aerodyn;servodyn;hydrodyn;subdyn;offshore"
    "5MW_ITIBarge_DLL_WTurb_WavesIrr"        , # "openfast;elastodyn;aerodyn;servodyn;hydrodyn;map;offshore"
    "5MW_TLP_DLL_WTurb_WavesIrr_WavesMulti"  , # "openfast;elastodyn;aerodyn;servodyn;hydrodyn;map;offshore"
    "5MW_OC3Spar_DLL_WTurb_WavesIrr"         , # "openfast;elastodyn;aerodyn;servodyn;hydrodyn;map;offshore"
    "5MW_OC4Semi_WSt_WavesWN"                , # "openfast;elastodyn;aerodyn;servodyn;hydrodyn;moordyn;offshore"
    "5MW_Land_BD_DLL_WTurb"                  , # "openfast;beamdyn;aerodyn;servodyn"
    "5MW_Land_BD_Init"                       , # "openfast;beamdyn;aerodyn;servodyn"
    "5MW_OC4Jckt_ExtPtfm"                    , # "openfast;elastodyn;extptfm"
    "HelicalWake_OLAF"                       , # "openfast;aerodyn;olaf"
    "EllipticalWing_OLAF"                    , # "openfast;aerodyn;olaf"
    "StC_test_OC4Semi"                       , # "openfast;servodyn;hydrodyn;moordyn;offshore;stc"
    "MHK_RM1_Fixed"                          , # "openfast;elastodyn;aerodyn;mhk"
    "MHK_RM1_Floating"                       , # "openfast;elastodyn;aerodyn;hydrodyn;moordyn;mhk"
    "MHK_RM1_Floating_wNacDrag"              , # "openfast;elastodyn;aerodyn;hydrodyn;moordyn;mhk"
    "Tailfin_FreeYaw1DOF_PolarBased"         , # "openfast;elastodyn;aerodyn"
    "Tailfin_FreeYaw1DOF_Unsteady"           , # "openfast;elastodyn;aerodyn"
    "5MW_Land_DLL_WTurb_ADsk"                , # "openfast;elastodyn;aerodisk"
    "5MW_Land_DLL_WTurb_ADsk_SED"            , # "openfast;simple-elastodyn;aerodisk"
    "5MW_Land_DLL_WTurb_SED"                 , # "openfast;simple-elastodyn;aerodyn"
]



def getPaths(OF_PATH = OF_PATH, REPOSITORY_ROOT = REPOSITORY_ROOT, BUILD_DIR = BUILD_DIR):

    """
    Function to get the paths for the OpenFAST executable, source directory, build directory, r-test directory, and test data directory

    Args:
    OF_PATH (str): Path to the OpenFAST executable
    REPOSITORY_ROOT (str): Path to the OpenFAST repository
    BUILD_DIR (str): Path to the build directory

    Returns:
    dict: Dictionary containing the paths
    """

    return {
        "executable": OF_PATH,
        "source_dir": REPOSITORY_ROOT,
        "build_dir": BUILD_DIR, # Location of the reg_tests directory inside the build directory created by the user
        "rtest_dir": osp.join(REPOSITORY_ROOT, "reg_tests", "r-test"),
        "test_data_dir": osp.join(REPOSITORY_ROOT, "reg_tests", "r-test", "glue-codes", "openfast"),
        "discon_dir": osp.join(BUILD_DIR, "glue-codes", "openfast", "5MW_Baseline", "ServoData"),
    }

def read_action(folder, path_dict = getPaths()):
    """
    Function to read the input deck for a given folder

    Args:
    folder (str): r-test folder name
    path_dict (dict): Dictionary containing the paths. Default is getPaths()

    Returns:
    dict: OpenFAST VariableTree 

    """
    print(f"Reading from {folder}")

    # Read input deck
    fast_reader = InputReader_OpenFAST()
    fast_reader.FAST_InputFile =  f'{folder}.fst'   # FAST input file (ext=.fst)
    fast_reader.FAST_directory = osp.join(path_dict['test_data_dir'], folder)   # Path to fst directory files
    fast_reader.execute()

    return fast_reader.fst_vt

def write_action(folder, fst_vt, path_dict = getPaths()):
    """
    Function to write the input deck for a given folder

    Args:
    folder (str): r-test folder name
    fst_vt (dict): OpenFAST VariableTree
    path_dict (dict): Dictionary containing the paths. Default is getPaths()

    """
    print(f"Writing to {folder}, with TMax = 2.0")

    # check if the folder exists, if not, mostly being called not from cmake, so create it
    if not osp.exists(osp.join(path_dict['build_dir'],'openfast_io')):
        Path(path_dict['build_dir']).mkdir(parents=True, exist_ok=True)

    fast_writer = InputWriter_OpenFAST()
    fast_writer.FAST_runDirectory = osp.join(path_dict['build_dir'],'openfast_io',folder)
    Path(fast_writer.FAST_runDirectory).mkdir(parents=True, exist_ok=True)
    fast_writer.FAST_namingOut    = folder

    fast_writer.fst_vt = dict(fst_vt)
    fst_vt = {}
    fst_vt['Fst', 'TMax'] = 2.
    fst_vt['Fst', 'TStart'] = 0.
    fst_vt['Fst','OutFileFmt'] = 3
    # check if the case needs ServoDyn
    if 'DLL_FileName' in fast_writer.fst_vt['ServoDyn']:
        # Copy the DISCON name and add the path to the build location
        fst_vt['ServoDyn', 'DLL_FileName'] = osp.join(path_dict['discon_dir'], osp.basename(fast_writer.fst_vt['ServoDyn']['DLL_FileName'])) 
    fast_writer.update(fst_update=fst_vt)
    fast_writer.execute()

def run_action(folder, path_dict = getPaths()):
    """
    Function to run the simulation for a given folder

    Args:
    folder (str): r-test folder name
    path_dict (dict): Dictionary containing the paths. Default is getPaths()

    """
    print(f"Running simulation for {folder}")
    command = [f"{path_dict['executable']}",  f"{osp.join(path_dict['build_dir'],'openfast_io', folder, f'{folder}.fst')}"]
    with open(osp.join(path_dict['build_dir'],'openfast_io', folder, f'{folder}.log'), 'w') as f:
        subprocess.run(command, check=True, stdout=f, stderr=subprocess.STDOUT)
        f.close()

def check_ascii_out(folder, path_dict = getPaths()):
    """
    Function to read the ASCII output for a given folder
    
    Args:
    folder (str): r-test folder name
    path_dict (dict): Dictionary containing the paths. Default is getPaths()
    """
    print(f"Checking ASCII output for {folder}")
    asciiOutput = osp.join(path_dict['build_dir'],'openfast_io', folder, f"{folder}.out")
    fast_outout = FASTOutputFile(filename=asciiOutput)

def check_binary_out(folder, path_dict = getPaths()):
    """
    Function to read the binary output for a given folder

    Args:
    folder (str): r-test folder name
    path_dict (dict): Dictionary containing the paths. Default is getPaths()
    """
    print(f"Checking binary output for {folder}")
    binaryOutput = osp.join(path_dict['build_dir'],'openfast_io', folder, f"{folder}.outb")
    fast_outout = FASTOutputFile(filename=binaryOutput)

def check_fst_vt_with_source(folder, path_dict = getPaths()):

    """
    Function to check the fst_vt with the source for a given folder

    Args:
    folder (str): r-test folder name
    path_dict (dict): Dictionary containing the paths. Default is getPaths()
    """

    print(f"Checking the fst_vt with the source for {folder}")

    # creating the two InputReader_OpenFAST objects
    fast_reader1 = InputReader_OpenFAST() #  for the source
    fast_reader1.FAST_InputFile =  f'{folder}.fst'   # FAST input file (ext=.fst)
    fast_reader1.FAST_directory = osp.join(path_dict['test_data_dir'], folder)   # Path to fst directory files
    fast_reader1.execute()

    fast_reader2 = InputReader_OpenFAST() #  for the build
    fast_reader2.FAST_InputFile =  f'{folder}.fst'   # FAST input file (ext=.fst)
    fast_reader2.FAST_directory = osp.join(path_dict['build_dir'],'openfast_io',folder)   # Path to fst directory files
    fast_reader2.execute()

    # List of acceptable differences
    acceptable_diff = [
                        'TMax', # TMax is updated in the write_action
                        'TStart', # TStart is updated in the write_action
                        'OutFileFmt', # OutFileFmt is updated in the write_action 
    ]


    # compare the two fst_vt
    diff = compare_fst_vt(fast_reader1.fst_vt, fast_reader2.fst_vt, ignoreVars = acceptable_diff,
                   removeFileRef=True, removeArrayProps=True, print_diff=True)
    
    if diff:
        pytest.fail(f"fst_vt for {folder} is not matching with the source")



# Begining of the test
def test_rtest_cloned(request):
    """
    Function to check if the r-tests are cloned properly

    Args:
    request (fixture): pytest request
    """

    REPOSITORY_ROOT = osp.join(request.config.getoption("--source_dir"))
    path_dict = getPaths(REPOSITORY_ROOT=REPOSITORY_ROOT)

    if check_rtest_cloned(path_dict['test_data_dir']):
        assert True, "R-tests cloned properly"
    else:# stop the test if the r-tests are not cloned properly
        print("R-tests not cloned properly")
        sys.exit(1)

def test_DLLs_exist(request):
    """
    Function to check if the DISCON.dll file exists

    Args:
    request (fixture): pytest request
    """

    path_dict = getPaths(OF_PATH=osp.join(request.config.getoption("--build_dir")))

    # Check if the DISCON.dll file exists
    DISCON_DLL = osp.join(path_dict['discon_dir'], "DISCON.dll")
    if osp.exists(DISCON_DLL):
        assert True, f"DISCON.dll found at {DISCON_DLL}"
    else: # stop the test if the DISCON.dll is not found
        print(f"DISCON.dll not found at {DISCON_DLL}. Please build with ''' make regression_test_controllers ''' and try again.")
        sys.exit(1)

def test_openfast_executable_exists(request):
    """
    Function to check if the OpenFAST executable exists

    Args:
    request (fixture): pytest request
    """

    path_dict = getPaths(OF_PATH=osp.join(request.config.getoption("--executable")))

    if osp.exists(path_dict['executable']):
        assert True, f"OpenFAST executable found at {path_dict['executable']}"
    else: # stop the test if the OpenFAST executable is not found
        print(f"OpenFAST executable not found at {path_dict['executable']}. Please build OpenFAST and try again.")
        sys.exit(1)



# Parameterize the test function to run for each folder and action
@pytest.mark.parametrize("folder", FOLDERS_TO_RUN)
def test_openfast_io_read_write_run_readOut_verify(folder, request):
    """
    Function to test the read, write, run, check ASCII, check binary, and check fst_vt for a given folder

    Args:
    folder (str): r-test folder name
    request (fixture): pytest request
    """

    path_dict = getPaths(OF_PATH=osp.join(request.config.getoption("--executable")), 
                         REPOSITORY_ROOT=osp.join(request.config.getoption("--source_dir")), 
                         BUILD_DIR=osp.join(request.config.getoption("--build_dir")))


    try:
        action_name = "read"
        fst_vt = read_action(folder, path_dict = path_dict)

        action_name = "write"
        write_action(folder, fst_vt, path_dict = path_dict)

        action_name = "run"
        run_action(folder, path_dict = path_dict)

        action_name = "check ASCII"
        check_ascii_out(folder, path_dict = path_dict)

        action_name = "check binary"
        check_binary_out(folder, path_dict = path_dict)

        action_name = "check fst_vt"
        check_fst_vt_with_source(folder, path_dict = path_dict)

    except Exception as e:
        pytest.fail(f"Action '{action_name}' for folder '{folder}' failed with exception: {e}")

def main():
    """
    Main function to run the test for all the folders
    """

    # Initialize any necessary setup here

    for folderMain in FOLDERS_TO_RUN:
        print(" ")
        print(f"Processing folder: {folderMain}")

        # Assuming read_action, write_action, run_action, check_ascii_out, check_binary_out, and check_fst_vt_with_source are defined elsewhere
        data = read_action(folderMain)
        write_action(folderMain, data)
        run_action(folderMain)
        check_ascii_out(folderMain)
        check_binary_out(folderMain)
        check_fst_vt_with_source(folderMain)
        print(f"Successfully processed folder: {folderMain}")

        folderMain = None
        data = None

if __name__ == "__main__":
    """
    Run the main function if the script is run directly or through VSCode debugger
    """

    main()


"""Read/Write TurbSim File

copied with love from the weio library: https://github.com/ebranlard/weio

"""
import pandas as pd
import numpy as np
import os
import struct
import time

try:
    from .file import File, EmptyFileError
except:
    EmptyFileError = type('EmptyFileError', (Exception,),{})
    File=dict

class TurbSimFile(File):
    """ 
    Read/write a TurbSim turbulence file (.bts). The object behaves as a dictionary.

    Main keys
    ---------
    - 'u': velocity field, shape (3 x nt x ny x nz)
    - 'y', 'z', 't': space and time coordinates 
    - 'dt', 'ID', 'info'
    - 'zTwr', 'uTwr': tower coordinates and field if present (3 x nt x nTwr)
    - 'zRef', 'uRef': height and velocity at a reference point (usually not hub)

    Main methods
    ------------
    - read, write, toDataFrame, keys
    - valuesAt, vertProfile, horizontalPlane, verticalPlane, closestPoint
    - fitPowerLaw
    - makePeriodic, checkPeriodic

    Examples
    --------

        ts = TurbSimFile('Turb.bts')
        print(ts.keys())
        print(ts['u'].shape)  
        u,v,w = ts.valuesAt(y=10.5, z=90)


    """

    @staticmethod
    def defaultExtensions():
        return ['.bts']

    @staticmethod
    def formatName():
        return 'TurbSim binary'

    def __init__(self, filename=None, **kwargs):
        self.filename = None
        if filename:
            self.read(filename, **kwargs)

    def read(self, filename=None, header_only=False, tdecimals=8):
        """ read BTS file, with field: 
                     u    (3 x nt x ny x nz)
                     uTwr (3 x nt x nTwr)
        """
        if filename:
            self.filename = filename
        if not self.filename:
            raise Exception('No filename provided')
        if not os.path.isfile(self.filename):
            raise OSError(2,'File not found:',self.filename)
        if os.stat(self.filename).st_size == 0:
            raise EmptyFileError('File is empty:',self.filename)

        scl = np.zeros(3, np.float32); off = np.zeros(3, np.float32)
        with open(self.filename, mode='rb') as f:            
            # Reading header info
            ID, nz, ny, nTwr, nt                      = struct.unpack('<h4l', f.read(2+4*4))
            dz, dy, dt, uHub, zHub, zBottom           = struct.unpack('<6f' , f.read(6*4)  )
            scl[0],off[0],scl[1],off[1],scl[2],off[2] = struct.unpack('<6f' , f.read(6*4))
            nChar, = struct.unpack('<l',  f.read(4))
            info = (f.read(nChar)).decode()
            # Reading turbulence field
            if not header_only: 
                u    = np.zeros((3,nt,ny,nz))
                uTwr = np.zeros((3,nt,nTwr))
                # For loop on time (acts as buffer reading, and only possible way when nTwr>0)
                for it in range(nt):
                    Buffer = np.frombuffer(f.read(2*3*ny*nz), dtype=np.int16).astype(np.float32).reshape([3, ny, nz], order='F')
                    u[:,it,:,:]=Buffer
                    Buffer = np.frombuffer(f.read(2*3*nTwr), dtype=np.int16).astype(np.float32).reshape([3, nTwr], order='F')
                    uTwr[:,it,:]=Buffer
                u -= off[:, None, None, None]
                u /= scl[:, None, None, None]
                self['u']    = u
                uTwr -= off[:, None, None]
                uTwr /= scl[:, None, None]
                self['uTwr'] = uTwr
        self['info'] = info
        self['ID']   = ID
        self['dt']   = np.round(dt, tdecimals) # dt is stored in single precision in the TurbSim output
        self['y']    = np.arange(ny)*dy 
        self['y']   -= np.mean(self['y']) # y always centered on 0
        self['z']    = np.arange(nz)*dz +zBottom
        self['t']    = np.round(np.arange(nt)*dt, tdecimals)
        self['zTwr'] =-np.arange(nTwr)*dz + zBottom
        self['zRef'] = zHub
        self['uRef'] = uHub

    def write(self, filename=None):
        """ 
        write a BTS file, using the following keys: 'u','z','y','t','uTwr'
                     u    (3 x nt x ny x nz)
                     uTwr (3 x nt x nTwr)
        """
        if filename:
            self.filename = filename
        if not self.filename:
            raise Exception('No filename provided')

        nDim, nt, ny, nz = self['u'].shape
        if 'uTwr' not in self.keys() :
            self['uTwr']=np.zeros((3,nt,0))
        if 'ID' not in self.keys() :
            self['ID']=7

        _, _, nTwr = self['uTwr'].shape
        tsTwr  = self['uTwr']
        ts     = self['u']
        intmin = -32768
        intrng = 65535
        off    = np.empty((3), dtype    = np.float32)
        scl    = np.empty((3), dtype    = np.float32)
        info = 'Generated by TurbSimFile on {:s}.'.format(time.strftime('%d-%b-%Y at %H:%M:%S', time.localtime()))
        # Calculate scaling, offsets and scaling data
        out    = np.empty(ts.shape, dtype=np.int16)
        outTwr = np.empty(tsTwr.shape, dtype=np.int16)
        for k in range(3):
            all_min, all_max = ts[k].min(), ts[k].max()
            if nTwr>0:
                all_min=min(all_min, tsTwr[k].min())
                all_max=max(all_max, tsTwr[k].max())
            if all_min == all_max:
                scl[k] = 1
            else:
                scl[k] = intrng / (all_max-all_min)
            off[k]    = intmin - scl[k] * all_min
            out[k]    = (ts[k]    * scl[k] + off[k]).astype(np.int16)
            outTwr[k] = (tsTwr[k] * scl[k] + off[k]).astype(np.int16)
        z0 = self['z'][0]
        dz = self['z'][1]- self['z'][0]
        dy = self['y'][1]- self['y'][0]
        dt = self['t'][1]- self['t'][0]

        # Providing estimates of uHub and zHub even if these fields are not used
        zHub,uHub, bHub = self.hubValues()

        with open(self.filename, mode='wb') as f:            
            f.write(struct.pack('<h4l', self['ID'], nz, ny, nTwr, nt))
            f.write(struct.pack('<6f', dz, dy, dt, uHub, zHub, z0)) # NOTE uHub, zHub maybe not used
            f.write(struct.pack('<6f', scl[0],off[0],scl[1],off[1],scl[2],off[2]))
            f.write(struct.pack('<l' , len(info)))
            f.write(info.encode())
            try:
                for it in np.arange(nt):
                    f.write(out[:,it,:,:].tobytes(order='F'))
                    f.write(outTwr[:,it,:].tobytes(order='F'))
            except:
                for it in np.arange(nt):
                    f.write(out[:,it,:,:].tostring(order='F'))
                    f.write(outTwr[:,it,:].tostring(order='F'))

    # --------------------------------------------------------------------------------}
    # --- Convenient properties (matching Mann Box interface as well)
    # --------------------------------------------------------------------------------{
    @property
    def z(self): return self['z'] # np.arange(nz)*dz +zBottom

    @property
    def y(self): return self['y'] # np.arange(ny)*dy  - np.mean( np.arange(ny)*dy )
 
    @property
    def t(self): return self['t'] # np.arange(nt)*dt

    # NOTE: it would be best to use dz and dy as given in the file to avoid numerical issues
    @property
    def dz(self): return self['z'][1]-self['z'][0]

    @property
    def dy(self): return self['y'][1]-self['y'][0]

    @property
    def dt(self): return self['t'][1]-self['t'][0]

    @property
    def nz(self): return len(self.z)

    @property
    def ny(self): return len(self.y)

    @property
    def nt(self): return len(self.t)

    # --------------------------------------------------------------------------------}
    # --- Extracting relevant "Line" data at one point
    # --------------------------------------------------------------------------------{
    def valuesAt(self, y, z, method='nearest'):
        """ return wind speed time series at a point """
        if method == 'nearest':
            iy, iz = self.closestPoint(y, z)
            u = self['u'][0,:,iy,iz]
            v = self['u'][1,:,iy,iz]
            w = self['u'][2,:,iy,iz]
        else:
            raise NotImplementedError()
        return u, v, w

    def closestPoint(self, y, z):
        iy = np.argmin(np.abs(self['y']-y))
        iz = np.argmin(np.abs(self['z']-z))
        return iy,iz

    def hubValues(self, zHub=None):
        if zHub is None:
            try:
                zHub=float(self['zRef'])
                bHub=True
            except:
                bHub=False
                iz = np.argmin(np.abs(self['z']-(self['z'][0]+self['z'][-1])/2))
                zHub = self['z'][iz]
        else:
            bHub=True
        try:
            uHub=float(self['uRef'])
        except:
            iz = np.argmin(np.abs(self['z']-zHub))
            iy = np.argmin(np.abs(self['y']-(self['y'][0]+self['y'][-1])/2))
            uHub = np.mean(self['u'][0,:,iy,iz])
        return zHub, uHub, bHub

    def midValues(self):
        iy,iz = self.iMid
        zMid = self['z'][iz]
        #yMid = self['y'][iy] # always 0
        uMid = np.mean(self['u'][0,:,iy,iz])
        return zMid, uMid

    @property
    def iMid(self):
        iy = np.argmin(np.abs(self['y']-(self['y'][0]+self['y'][-1])/2))
        iz = np.argmin(np.abs(self['z']-(self['z'][0]+self['z'][-1])/2))
        return iy,iz

    def closestPoint(self, y, z):
        iy = np.argmin(np.abs(self['y']-y))
        iz = np.argmin(np.abs(self['z']-z))
        return iy,iz

    def _longiline(ts, iy0=None, iz0=None, removeMean=False):
        """ return velocity components on a longitudinal line
        If no index is provided, computed at mid box 
        """
        if iy0 is None:
            iy0,iz0 = ts.iMid
        u = ts['u'][0,:,iy0,iz0].copy()
        v = ts['u'][1,:,iy0,iz0].copy()
        w = ts['u'][2,:,iy0,iz0].copy()
        if removeMean:
            u -= np.mean(u)
            v -= np.mean(v)
            w -= np.mean(w)
        return u, v, w

    def _latline(ts, ix0=None, iz0=None, removeMean=False):
        """ return velocity components on a lateral line
        If no index is provided, computed at mid box 
        """
        if ix0 is None:
            iy0,iz0 = ts.iMid
            ix0=int(len(ts['t'])/2)
        u = ts['u'][0,ix0,:,iz0].copy()
        v = ts['u'][1,ix0,:,iz0].copy()
        w = ts['u'][2,ix0,:,iz0].copy()
        if removeMean:
            u -= np.mean(u)
            v -= np.mean(v)
            w -= np.mean(w)
        return u, v, w

    def _vertline(ts, ix0=None, iy0=None, removeMean=False):
        """ return velocity components on a vertical line
        If no index is provided, computed at mid box 
        """
        if ix0 is None:
            iy0,iz0 = ts.iMid
            ix0=int(len(ts['t'])/2)
        u = ts['u'][0,ix0,iy0,:].copy()
        v = ts['u'][1,ix0,iy0,:].copy()
        w = ts['u'][2,ix0,iy0,:].copy()
        if removeMean:
            u -= np.mean(u)
            v -= np.mean(v)
            w -= np.mean(w)
        return u, v, w

    # --------------------------------------------------------------------------------}
    # --- Extracting plane data at one point
    # --------------------------------------------------------------------------------{
    def horizontalPlane(ts, z=None, iz0=None, removeMean=False):
        """ return velocity components on a horizontal plane z=cst, (time x ny)
        If no z value is provided, returned at mid box 
        """
        if z is None and iz0 is None:
            _,iz0 = ts.iMid
        elif z is not None:
            _, iz0 = ts.closestPoint(ts.y[0], z) 

        u = ts['u'][0,:,:,iz0].copy()
        v = ts['u'][1,:,:,iz0].copy()
        w = ts['u'][2,:,:,iz0].copy()
        if removeMean:
            u -= np.mean(u)
            v -= np.mean(v)
            w -= np.mean(w)
        return u, v, w

    def verticalPlane(ts, y=None, iy0=None, removeMean=False):
        """ return velocity components on a vertical plane y=cst, (time x nz)
        If no y value is provided, returned at mid box 
        """
        if y is None and iy0 is None:
            iy0,_ = ts.iMid
        elif y is not None:
            iy0, _ = ts.closestPoint(y, ts.z[0]) 

        u = ts['u'][0,:,iy0,:].copy()
        v = ts['u'][1,:,iy0,:].copy()
        w = ts['u'][2,:,iy0,:].copy()
        if removeMean:
            u -= np.mean(u)
            v -= np.mean(v)
            w -= np.mean(w)
        return u, v, w

    def normalPlane(ts, t=None, it0=None, removeMean=False):
        """ return velocity components on a normal plane t=cst, (ny x nz)
        """
        if t is None and it0 is None:
            it0 = 0
        elif t is not None:
            it0 = np.argmin(np.abs(self['t']-t))

        u = ts['u'][0,it0,:,:].copy()
        v = ts['u'][1,it0,:,:].copy()
        w = ts['u'][2,it0,:,:].copy()
        if removeMean:
            u -= np.mean(u)
            v -= np.mean(v)
            w -= np.mean(w)
        return u, v, w


    # --------------------------------------------------------------------------------}
    # --- Extracting average data
    # --------------------------------------------------------------------------------{
    def vertProfile(self, y_span='full'):
        """ Vertical profile of the box
        INPUTS:
         - y_span: if 'full', average the vertical profile across all y-values
                   if 'mid', average the vertical profile at the middle y value
        """
        if y_span=='full':
            # Compute statistics with respect to time first, then average over "y"
            m = np.mean(np.mean(self['u'][:,:,:,:], axis=1), axis=1)
            s = np.mean(np.std( self['u'][:,:,:,:], axis=1), axis=1)
        elif y_span=='mid':
            iy, iz = self.iMid
            m = np.mean(self['u'][:,:,iy,:], axis=1)
            s = np.std( self['u'][:,:,iy,:], axis=1)
        else:
            raise NotImplementedError()
        return self.z, m, s


    # --------------------------------------------------------------------------------}
    # --- Computation of useful quantities
    # --------------------------------------------------------------------------------{
    def crosscorr_y(ts, iy0=None, iz0=None):
        """ Cross correlation along y
        If no index is provided, computed at mid box 
        """
        y = ts['y']
        if iy0 is None:
            iy0,iz0 = ts.iMid
        u, v, w = ts._longiline(iy0=iy0, iz0=iz0, removeMean=True)
        rho_uu_y=np.zeros(len(y))
        rho_vv_y=np.zeros(len(y))
        rho_ww_y=np.zeros(len(y))
        for iy,_ in enumerate(y):
            ud, vd, wd = ts._longiline(iy0=iy, iz0=iz0, removeMean=True)
            rho_uu_y[iy] = np.mean(u*ud)/(np.std(u)*np.std(ud))
            rho_vv_y[iy] = np.mean(v*vd)/(np.std(v)*np.std(vd))
            rho_ww_y[iy] = np.mean(w*wd)/(np.std(w)*np.std(wd))
        return y, rho_uu_y, rho_vv_y, rho_ww_y

    def crosscorr_z(ts, iy0=None, iz0=None):
        """ 
        Cross correlation along z, mid box
        If no index is provided, computed at mid box 
        """
        z = ts['z']
        if iy0 is None:
            iy0,iz0 = ts.iMid
        u, v, w = ts._longiline(iy0=iy0, iz0=iz0, removeMean=True)
        rho_uu_z = np.zeros(len(z))
        rho_vv_z = np.zeros(len(z))
        rho_ww_z = np.zeros(len(z))
        for iz,_ in enumerate(z):
            ud, vd, wd = ts._longiline(iy0=iy0, iz0=iz, removeMean=True)
            rho_uu_z[iz] = np.mean(u*ud)/(np.std(u)*np.std(ud))
            rho_vv_z[iz] = np.mean(v*vd)/(np.std(v)*np.std(vd))
            rho_ww_z[iz] = np.mean(w*wd)/(np.std(w)*np.std(wd))
        return z, rho_uu_z, rho_vv_z, rho_ww_z


    def csd_longi(ts, iy0=None, iz0=None):
        """ Compute cross spectral density
        If no index is provided, computed at mid box 
        """
        import scipy.signal as sig
        u, v, w = ts._longiline(iy0=iy0, iz0=iz0, removeMean=True)
        t       = ts['t']
        dt      = t[1]-t[0]
        fs      = 1/dt
        fc, chi_uu = sig.csd(u, u, fs=fs, scaling='density') #nperseg=4096, noverlap=2048, detrend='constant')
        fc, chi_vv = sig.csd(v, v, fs=fs, scaling='density') #nperseg=4096, noverlap=2048, detrend='constant')
        fc, chi_ww = sig.csd(w, w, fs=fs, scaling='density') #nperseg=4096, noverlap=2048, detrend='constant')
        return fc, chi_uu, chi_vv, chi_ww

    def csd_lat(ts, ix0=None, iz0=None):
        """ Compute lateral cross spectral density
        If no index is provided, computed at mid box 
        """
        try:
            import scipy.signal as sig
        except:
            import pydatview.tools.spectral as sig
        u, v, w = ts._latline(ix0=ix0, iz0=iz0, removeMean=True)
        t       = ts['t']
        dt      = t[1]-t[0]
        fs      = 1/dt
        fc, chi_uu = sig.csd(u, u, fs=fs, scaling='density') #nperseg=4096, noverlap=2048, detrend='constant')
        fc, chi_vv = sig.csd(v, v, fs=fs, scaling='density') #nperseg=4096, noverlap=2048, detrend='constant')
        fc, chi_ww = sig.csd(w, w, fs=fs, scaling='density') #nperseg=4096, noverlap=2048, detrend='constant')
        return fc, chi_uu, chi_vv, chi_ww

    def csd_vert(ts, ix0=None, iy0=None):
        """ Compute vertical cross spectral density
        If no index is provided, computed at mid box 
        """
        try:
            import scipy.signal as sig
        except:
            import pydatview.tools.spectral as sig
        t       = ts['t']
        dt      = t[1]-t[0]
        fs      = 1/dt
        u, v, w = ts._vertline(ix0=ix0, iy0=iy0, removeMean=True)
        u= u-np.mean(u)
        v= v-np.mean(v)
        w= w-np.mean(w)
        fc, chi_uu = sig.csd(u, u, fs=fs, scaling='density') #nperseg=4096, noverlap=2048, detrend='constant')
        fc, chi_vv = sig.csd(v, v, fs=fs, scaling='density') #nperseg=4096, noverlap=2048, detrend='constant')
        fc, chi_ww = sig.csd(w, w, fs=fs, scaling='density') #nperseg=4096, noverlap=2048, detrend='constant')
        return fc, chi_uu, chi_vv, chi_ww


    def coherence_longi(ts, iy0=None, iz0=None):
        """ Coherence on a longitudinal line for different delta y and delta z
        compared to a given point with index iy0,iz0
        """
        try:
            import scipy.signal as sig
        except:
            import pydatview.tools.spectral as sig
        if iy0 is None:
            iy0,iz0 = ts.iMid
        u, v, w = ts._longiline(iy0=iy0, iz0=iz0, removeMean=True)
        y = ts['y']
        z = ts['z']
        diy=1
        dy=y[iy]-y[iy0]
        # TODO
        iy = iy0+diy
        ud, vd, wd = ts._longiline(iy0=iy, iz0=iz0, removeMean=True)
        fc, coh_uu_y1 = sig.coherence(u,ud, fs=fs)
        _ , coh_vv_y1 = sig.coherence(v,vd, fs=fs)
        _ , coh_ww_y1 = sig.coherence(w,wd, fs=fs)

        iy = iy+diy
        ud, vd, wd = ts._longiline(iy0=iy, iz0=iz0, removeMean=False)
        _ , coh_uu_y2 = sig.coherence(u,ud, fs=fs)
        _ , coh_vv_y2 = sig.coherence(v,vd, fs=fs)
        _ , coh_ww_y2 = sig.coherence(w,wd, fs=fs)


    # --------------------------------------------------------------------------------}
    # --- Modifierss
    # --------------------------------------------------------------------------------{
    def scale(self, new_mean=None, new_std=None, component=0, reference='mid', y_ref=0, z_ref=None):
        """ 
        TODO needs more thinking
        """
        # mean/std values for each points in the plane (averaged with time)
        old_plane_mean = np.mean(self['u'][component,:,:,:],axis=0)
        old_plane_std  = np.std( self['u'][component,:,:,:],axis=0)
        if reference=='mid':
            iy,iz = self.iMid
            old_mean = np.mean(self['u'][component,:,iy,iz])
            old_std  = np.std (self['u'][component,:,iy,iz])
        elif reference=='point':
            iy, iz = self.closestPoint(y_ref, z_ref)
            old_mean = np.mean(self['u'][component,:,iy,iz])
            old_std  = np.std (self['u'][component,:,iy,iz])
        else:
            raise NotImplementedError(reference)
        # Scaling standard deviation without affecting the mean
        self['u'][component,:,:,:] -= old_plane_mean
        if new_std is not None:
            self['u'][component,:,:,:] *= new_std/old_std
        self['u'][component,:,:,:] += old_plane_mean

        # Scaling mean
        if new_mean is not None:
            self['u'][component,:,:,:] += -old_mean+new_mean

        # Sanity check
        new_mean2= np.mean(self['u'][component,:,iy,iz])
        new_std2 = np.std(self['u'][component,:,iy,iz])
        if new_mean is not None:
            print('New mean: {:7.3f}  (target: {:7.3f}, old: {:7.3f})'.format(new_mean2, new_mean, old_mean))
        if new_std is not None:
            print('New std : {:7.3f}  (target: {:7.3f}, old: {:7.3f})'.format(new_std2 , new_std , old_std))

    def makePeriodic(self):
        """ Make the box periodic in the streamwise direction by mirroring it """
        nDim, nt0, ny, nz = self['u'].shape
        u = self['u'].copy()
        del self['u']

        nt = 2*len(self['t'])-2
        dt = self['t'][1]- self['t'][0]
        self['u']  = np.zeros((nDim,nt,ny,nz))
        self['u'][:,:nt0,:,:] = u
        self['u'][:,nt0:,:,:] = np.flip(u[:,1:-1,:,:],axis=1)
        self['t'] = np.arange(nt)*dt
        if 'uTwr' in self.keys():
            _, _, nTwr = self['uTwr'].shape
            uTwr = self['uTwr'].copy()
            del self['uTwr']
            # empty tower for now
            self['uTwr'] = np.zeros((nDim,nt,nTwr))
            self['uTwr'][:,:nt0,:] = uTwr
            self['uTwr'][:,nt0:,:] = np.flip(uTwr[:,1:-1,:],axis=1)

        self['ID']=8 # Periodic


    def checkPeriodic(self, sigmaTol=1.5, aTol=0.5):
        """ Check periodicity in u """
        ic=0
        sig  = np.std(self['u'][ic,:,:,:],axis=0)
        mean = np.mean(self['u'][ic,:,:,:],axis=0)
        u_first= self['u'][ic,0 ,:,:]
        u_last = self['u'][ic,-1,:,:]
        relSig = np.abs(u_first-u_last)/sig
        compPeriodic = (np.max(relSig) < sigmaTol) and (np.mean(np.abs(u_first-u_last))<aTol)
        return compPeriodic


    def __repr__(self):
        s='<{} object> with keys:\n'.format(type(self).__name__)
        s+=' - filename: {}\n'.format(self.filename)
        s+=' - ID: {}\n'.format(self['ID'])
        s+=' - z: [{} ... {}],  dz: {}, n: {} \n'.format(self['z'][0],self['z'][-1],self['z'][1]-self['z'][0],len(self['z']))
        s+=' - y: [{} ... {}],  dy: {}, n: {} \n'.format(self['y'][0],self['y'][-1],self['y'][1]-self['y'][0],len(self['y']))
        s+=' - t: [{} ... {}],  dt: {}, n: {} \n'.format(self['t'][0],self['t'][-1],self['t'][1]-self['t'][0],len(self['t']))
        if 'u' in self.keys():
            s+=' - u: ({} x {} x {} x {}) \n'.format(*(self['u'].shape))
            ux,uy,uz=self['u'][0], self['u'][1], self['u'][2]
            s+='    ux: min: {}, max: {}, mean: {} \n'.format(np.min(ux), np.max(ux), np.mean(ux))
            s+='    uy: min: {}, max: {}, mean: {} \n'.format(np.min(uy), np.max(uy), np.mean(uy))
            s+='    uz: min: {}, max: {}, mean: {} \n'.format(np.min(uz), np.max(uz), np.mean(uz))
            # Mid of box, nearest neighbor
            iy,iz = self.iMid
            zMid=self['z'][iz]
            yMid=self['y'][iy]
            uMid = np.mean(self['u'][0,:,iy,iz])
            s+='    yMid: {} - zMid: {} - iy: {} - iz: {} - uMid: {} (nearest neighbor))\n'.format(yMid, zMid, iy, iz, uMid)
#         zMid, uMid, bHub = self.hubValues()
#         if bHub:
#             s+='    z"Hub": {} - u"Hub": {} (NOTE: values at TurbSim "hub")\n'.format(zMid, uMid)

        # Tower
        if 'zTwr' in self.keys() and len(self['zTwr'])>0:
            s+=' - zTwr: [{} ... {}],  dz: {}, n: {} \n'.format(self['zTwr'][0],self['zTwr'][-1],self['zTwr'][1]-self['zTwr'][0],len(self['zTwr']))
        if 'uTwr' in self.keys() and self['uTwr'].shape[2]>0:
            s+=' - uTwr: ({} x {} x {} ) \n'.format(*(self['uTwr'].shape))
            ux,uy,uz=self['uTwr'][0], self['uTwr'][1], self['uTwr'][2]
            s+='    ux: min: {}, max: {}, mean: {} \n'.format(np.min(ux), np.max(ux), np.mean(ux))
            s+='    uy: min: {}, max: {}, mean: {} \n'.format(np.min(uy), np.max(uy), np.mean(uy))
            s+='    uz: min: {}, max: {}, mean: {} \n'.format(np.min(uz), np.max(uz), np.mean(uz))
        s += ' Useful methods:\n'
        s += ' - read, write, toDataFrame, keys\n'
        s += ' - valuesAt, vertProfile, horizontalPlane, verticalPlane, closestPoint\n'
        s += ' - fitPowerLaw\n'
        s += ' - makePeriodic, checkPeriodic\n'
        return s

    def toDataFrame(self):
        dfs={}

        ny = len(self['y'])
        nz = len(self['y'])
        # Index at mid box
        iy,iz = self.iMid

        # Mean vertical profile
        z, m, s = self.vertProfile()
        ti = s/m*100
        cols=['z_[m]','u_[m/s]','v_[m/s]','w_[m/s]','sigma_u_[m/s]','sigma_v_[m/s]','sigma_w_[m/s]','TI_[%]']
        data = np.column_stack((z, m[0,:],m[1,:],m[2,:],s[0,:],s[1,:],s[2,:],ti[0,:]))
        dfs['VertProfile'] = pd.DataFrame(data = data ,columns = cols)

        # Mid time series
        u = self['u'][:,:,iy,iz]
        cols=['t_[s]','u_[m/s]','v_[m/s]','w_[m/s]']
        data = np.column_stack((self['t'],u[0,:],u[1,:],u[2,:]))
        dfs['ZMidLine'] = pd.DataFrame(data = data ,columns = cols)


        # ZMid YStart time series
        u = self['u'][:,:,0,iz]
        cols=['t_[s]','u_[m/s]','v_[m/s]','w_[m/s]']
        data = np.column_stack((self['t'],u[0,:],u[1,:],u[2,:]))
        dfs['ZMidYStartLine'] = pd.DataFrame(data = data ,columns = cols)

        # ZMid YEnd time series
        u = self['u'][:,:,-1,iz]
        cols=['t_[s]','u_[m/s]','v_[m/s]','w_[m/s]']
        data = np.column_stack((self['t'],u[0,:],u[1,:],u[2,:]))
        dfs['ZMidYEndLine'] = pd.DataFrame(data = data ,columns = cols)

        # Mid crosscorr y
        y, rho_uu_y, rho_vv_y, rho_ww_y = self.crosscorr_y()
        cols = ['y_[m]', 'rho_uu_[-]','rho_vv_[-]','rho_ww_[-]']
        data = np.column_stack((y, rho_uu_y, rho_vv_y, rho_ww_y))
        dfs['Mid_xcorr_y'] = pd.DataFrame(data = data ,columns = cols)

        # Mid crosscorr z
        z, rho_uu_z, rho_vv_z, rho_ww_z = self.crosscorr_z()
        cols = ['z_[m]', 'rho_uu_[-]','rho_vv_[-]','rho_ww_[-]']
        data = np.column_stack((z, rho_uu_z, rho_vv_z, rho_ww_z))
        dfs['Mid_xcorr_z'] = pd.DataFrame(data = data ,columns = cols)

        # Mid csd
        try:
            import warnings
            with warnings.catch_warnings():
                warnings.filterwarnings('ignore') #, category=DeprecationWarning)
            fc, chi_uu, chi_vv, chi_ww = self.csd_longi()
            cols = ['f_[Hz]','chi_uu_[-]', 'chi_vv_[-]','chi_ww_[-]']
            data = np.column_stack((fc, chi_uu, chi_vv, chi_ww))
            dfs['Mid_csd_longi'] = pd.DataFrame(data = data ,columns = cols)

            # Mid csd
            fc, chi_uu, chi_vv, chi_ww = self.csd_lat()
            cols = ['f_[Hz]','chi_uu_[-]', 'chi_vv_[-]','chi_ww_[-]']
            data = np.column_stack((fc, chi_uu, chi_vv, chi_ww))
            dfs['Mid_csd_lat'] = pd.DataFrame(data = data ,columns = cols)

            # Mid csd
            fc, chi_uu, chi_vv, chi_ww = self.csd_vert()
            cols = ['f_[Hz]','chi_uu_[-]', 'chi_vv_[-]','chi_ww_[-]']
            data = np.column_stack((fc, chi_uu, chi_vv, chi_ww))
            dfs['Mid_csd_vert'] = pd.DataFrame(data = data ,columns = cols)
        except ModuleNotFoundError:
            print('Module scipy.signal not available')
        except ImportError:
            print('Likely issue with fftpack')


        # Hub time series
        #try:
        #    zHub = self['zHub']
        #    iz = np.argmin(np.abs(self['z']-zHub))
        #    u = self['u'][:,:,iy,iz]
        #    Cols=['t_[s]','u_[m/s]','v_[m/s]','w_[m/s]']
        #    data = np.column_stack((self['t'],u[0,:],u[1,:],u[2,:]))
        #    dfs['TSHubLine'] = pd.DataFrame(data = data ,columns = Cols)
        #except:
        #    pass
        return dfs

    def to2DFields(self, nTOut=10, nYOut=3, nZOut=3, **kwargs):
        import xarray as xr
        if len(kwargs.keys())>0:
            print('[WARN] TurbSimFile: to2DFields: ignored keys: ',kwargs.keys())
        # Sanity check
        nTOut = int(nTOut)
        nYOut = int(nYOut)
        nZOut = int(nZOut)
        # 'u': velocity field, shape (3 x nt x ny x nz)
        IT = list(np.arange(nTOut)) +  [len(self.t)-1]
        IT = np.unique(IT)

        IY = np.unique(np.linspace(0,len(self['y'])-1, nYOut).astype(int))
        IZ = np.unique(np.linspace(0,len(self['z'])-1, nZOut).astype(int))
        Fields=[]
        # --- YZ planes
        s1 = 'TSR'
        s2 = 'pitch'
        ds = xr.Dataset(coords={'t': self.t, 'y': self.y, 'z': self.z})
        ds['t'].attrs['unit'] = 's'
        ds['y'].attrs['unit'] = 'm'
        ds['z'].attrs['unit'] = 'm'

        ds['(y,z)_u_avg_[m/s]']= (['y','z'], np.squeeze(np.mean(self['u'][0,:,:,:], axis=0)))
        ds['(y,z)_v_avg_[m/s]']= (['y','z'], np.squeeze(np.mean(self['u'][1,:,:,:], axis=0)))
        ds['(y,z)_w_avg_[m/s]']= (['y','z'], np.squeeze(np.mean(self['u'][2,:,:,:], axis=0)))

        ds['(t,y)_u_avg_[m/s]']= (['t','y'], np.squeeze(np.mean(self['u'][0,:,:,:], axis=2)))
        ds['(t,y)_v_avg_[m/s]']= (['t','y'], np.squeeze(np.mean(self['u'][1,:,:,:], axis=2)))
        ds['(t,y)_w_avg_[m/s]']= (['t','y'], np.squeeze(np.mean(self['u'][2,:,:,:], axis=2)))

        ds['(t,z)_u_avg_[m/s]']= (['t','z'], np.squeeze(np.mean(self['u'][0,:,:,:], axis=1)))
        ds['(t,z)_v_avg_[m/s]']= (['t','z'], np.squeeze(np.mean(self['u'][1,:,:,:], axis=1)))
        ds['(t,z)_w_avg_[m/s]']= (['t','z'], np.squeeze(np.mean(self['u'][2,:,:,:], axis=1)))

        for it in IT:
            ds['(y,z)_u_t={:.2f}_[m/s]'.format(self.t[it])] = (['y','z'], np.squeeze(self['u'][0,it,:,:]))
        for it in IT:
            ds['(y,z)_v_t={:.2f}_[m/s]'.format(self.t[it])] = (['y','z'], np.squeeze(self['u'][1,it,:,:]))
        for it in IT:
            ds['(y,z)_w_t={:.2f}_[m/s]'.format(self.t[it])] = (['y','z'], np.squeeze(self['u'][2,it,:,:]))
        # --- TZ planes
        for iy in IY:
            ds['(t,z)_u_y={:.2f}_[m/s]'.format(self['y'][iy])] = (['t','z'], np.squeeze(self['u'][0,:,iy,:]))
        for iy in IY:
            ds['(t,z)_v_y={:.2f}_[m/s]'.format(self['y'][iy])] = (['t','z'], np.squeeze(self['u'][1,:,iy,:]))
        for iy in IY:
            ds['(t,z)_w_y={:.2f}_[m/s]'.format(self['y'][iy])] = (['t','z'], np.squeeze(self['u'][2,:,iy,:]))
        # --- TY planes
        for iz in IZ:
            ds['(t,y)_u_z={:.2f}_[m/s]'.format(self['z'][iz])] = (['t','y'], np.squeeze(self['u'][0,:,:,iz]))
        for iz in IZ:
            ds['(t,y)_v_z={:.2f}_[m/s]'.format(self['z'][iz])] = (['t','y'], np.squeeze(self['u'][1,:,:,iz]))
        for iz in IZ:
            ds['(t,y)_w_z={:.2f}_[m/s]'.format(self['z'][iz])] = (['t','y'], np.squeeze(self['u'][2,:,:,iz]))
        return ds

    def toDataset(self):
        """
        Convert the data that was read in into a xarray Dataset
        
        # TODO SORT OUT THE DIFFERENCE WITH toDataSet
        """
        from xarray import IndexVariable, DataArray, Dataset
        
        print('[TODO] pyFAST.input_output.turbsim_file.toDataset: merge with function toDataSet')

        y      = IndexVariable("y", self.y, attrs={"description":"lateral coordinate","units":"m"})
        zround = np.asarray([np.round(zz,6) for zz in self.z]) #the open function here returns something like *.0000000001 which is annoying
        z      = IndexVariable("z", zround, attrs={"description":"vertical coordinate","units":"m"})
        time   = IndexVariable("time", self.t, attrs={"description":"time since start of simulation","units":"s"})

        da = {}
        for component,direction,velname in zip([0,1,2],["x","y","z"],["u","v","w"]):
            # the dataset produced here has y/z axes swapped relative to data stored in original object
            velocity = np.swapaxes(self["u"][component,...],1,2)     
            da[velname] = DataArray(velocity, 
                         coords={"time":time,"y":y,"z":z}, 
                         dims=["time","y","z"], 
                         name="velocity", 
                         attrs={"description":"velocity along {0}".format(direction),"units":"m/s"})

        return Dataset(data_vars=da, coords={"time":time,"y":y,"z":z})      

    def toDataSet(self, datetime=False):
        """
        Convert the data that was read in into a xarray Dataset
        
        # TODO SORT OUT THE DIFFERENCE WITH toDataset
        """
        import xarray as xr
        
        print('[TODO] pyFAST.input_output.turbsim_file.toDataSet: should be discontinued')        
        print('[TODO] pyFAST.input_output.turbsim_file.toDataSet: merge with function toDataset')        

        if datetime:
            timearray = pd.to_datetime(self['t'], unit='s', origin=pd.to_datetime('2000-01-01 00:00:00'))
            timestr   = 'datetime'
        else:
            timearray = self['t']
            timestr   = 'time'

        ds = xr.Dataset(
            data_vars=dict(
                u=([timestr,'y','z'], self['u'][0,:,:,:]),
                v=([timestr,'y','z'], self['u'][1,:,:,:]),
                w=([timestr,'y','z'], self['u'][2,:,:,:]),
            ),
            coords={
                timestr : timearray,
                'y' : self['y'],
                'z' : self['z'],
            },
        )

        # Add mean computations
        ds['up'] = ds['u'] - ds['u'].mean(dim=timestr)
        ds['vp'] = ds['v'] - ds['v'].mean(dim=timestr)
        ds['wp'] = ds['w'] - ds['w'].mean(dim=timestr)

        if datetime:
            # Add time (in s) to the variable list
            ds['time'] = (('datetime'), self['t'])

        return ds

    # Useful converters
    def fromAMRWind(self, filename, timestep, output_frequency, sampling_identifier, verbose=1, fileout=None, zref=None, xloc=None):
        """
        Reads a AMRWind netcdf file, grabs a group of sampling planes (e.g. p_slice),
          return an instance of TurbSimFile, optionally write turbsim file to disk


        Parameters
        ----------
        filename : str,
            full path to netcdf file generated by amrwind
        timestep : float,
            amr-wind code timestep (time.fixed_dt)
        output_frequency : int,
            frequency chosen for sampling output in amrwind input file (sampling.output_frequency)
        sampling_identifier : str,
            identifier of the sampling being requested (an entry of sampling.labels in amrwind input file)
        zref : float,
            height to be written to turbsim as the reference height. if none is given, it is taken as the vertical centerpoint of the slice
        """
        try:
            from pyFAST.input_output.amrwind_file import AMRWindFile
        except:
            try:
                from .amrwind_file import AMRWindFile
            except:
                from amrwind_file import AMRWindFile
                
        obj = AMRWindFile(filename,timestep,output_frequency, group_name=sampling_identifier)

        self["u"]          = np.ndarray((3,obj.nt,obj.ny,obj.nz)) 
        
        xloc = float(obj.data.x[0]) if xloc is None else xloc
        if verbose:
            print("Grabbing the slice at x={0} m".format(xloc))
        self['u'][0,:,:,:] = np.swapaxes(obj.data.u.sel(x=xloc).values,1,2)
        self['u'][1,:,:,:] = np.swapaxes(obj.data.v.sel(x=xloc).values,1,2)
        self['u'][2,:,:,:] = np.swapaxes(obj.data.w.sel(x=xloc).values,1,2)        
        self['t']          = obj.data.t.values

        self['y']  = obj.data.y.values
        self['z']  = obj.data.z.values
        self['dt'] = obj.output_dt
        
        self['ID'] = 7 
        ltime = time.strftime('%d-%b-%Y at %H:%M:%S', time.localtime())
        self['info'] = 'Converted from AMRWind output file {0} {1:s}.'.format(filename,ltime)
        
        iz           = int(obj.nz/2)
        self['zRef'] = float(obj.data.z[iz]) if zref is None else zref
        if verbose:
            print("Setting the TurbSim file reference height to z={0} m".format(self["zRef"]))                

        self['uRef'] = float(obj.data.u.sel(x=xloc).sel(y=0).sel(z=self["zRef"]).mean().values)
        self['zRef'], self['uRef'], bHub = self.hubValues()
        
        if fileout is not None:
            filebase = os.path.splitext(filename)[1]
            fileout = filebase+".bts"
            if verbose:
                print("===> {0}".format(fileout))
            self.write(fileout) 
    
    def fromAMRWind_legacy(self, filename, dt, nt, y, z, sampling_identifier='p_sw2'):
        """
        Convert current TurbSim file into one generated from AMR-Wind LES sampling data in .nc format
        Assumes:
          --  u, v, w (nt, nx * ny * nz)
          --  u is aligned with x-axis (flow is not rotated) - this consideration needs to be added


        INPUTS:
          - filename: (string) full path to .nc sampling data file
          - sampling_identifier: (string) name of sampling plane group from .inp file (e.g. "p_sw2")
          - dt: timestep size [s]
          - nt: number of timesteps (sequential) you want to read in, starting at the first timestep available
        INPUTS: TODO
          - y: user-defined vector of coordinate positions in y
          - z: user-defined vector of coordinate positions in z
          - uref: (float) reference mean velocity (e.g. 8.0 hub height mean velocity from input file)
          - zref: (float) hub height (e.t. 150.0)
        """
        import xarray as xr

        print('[TODO] fromAMRWind_legacy: function might be unfinished. Merge with fromAMRWind')
        print('[TODO] fromAMRWind_legacy: figure out y, and z from data (see fromAMRWind)')
        
        # read in sampling data plane
        ds = xr.open_dataset(filename,
                              engine='netcdf4',
                              group=sampling_identifier)
        ny, nz, _ = ds.attrs['ijk_dims']
        noffsets  = len(ds.attrs['offsets'])
        t         = np.arange(0, dt*(nt-0.5), dt)
        print('max time [s] = ', t[-1])

        self['u']=np.ndarray((3,nt,ny,nz)) #, buffer=shm.buf)
        # read in AMRWind velocity data
        self['u'][0,:,:,:] = ds['velocityx'].isel(num_time_steps=slice(0,nt)).values.reshape(nt,noffsets,ny,nz)[:,1,:,:] # last index = 1 refers to 2nd offset plane at -1200 m
        self['u'][1,:,:,:] = ds['velocityy'].isel(num_time_steps=slice(0,nt)).values.reshape(nt,noffsets,ny,nz)[:,1,:,:]
        self['u'][2,:,:,:] = ds['velocityz'].isel(num_time_steps=slice(0,nt)).values.reshape(nt,noffsets,ny,nz)[:,1,:,:]
        self['t']  = t
        self['y']  = y
        self['z']  = z
        self['dt'] = dt
        # TODO
        self['ID'] = 7 # ...
        self['info'] = 'Converted from AMRWind fields {:s}.'.format(time.strftime('%d-%b-%Y at %H:%M:%S', time.localtime()))
#         self['zTwr'] = np.array([])
#         self['uTwr'] = np.array([])
        self['zRef'] = zref #None
        self['uRef'] = uref #None
        self['zRef'], self['uRef'], bHub = self.hubValues()

    def fromMannBox(self, u, v, w, dx, U, y, z, addU=None):
        """ 
        Convert current TurbSim file into one generated from MannBox
        Assumes: 
             u, v, w (nt x ny x nz)

             y: goes from -ly/2 to ly/2  this is an IMPORTANT subtlety
                The field u needs to respect this convention!
                (fields from weio.mannbox_file do respect this convention
                but when exported to binary files, the y axis is flipped again)
        
        INPUTS:
          - u, v, w : mann box fields
          - dx: axial spacing of mann box (to compute time)
          - U: reference speed of mann box (to compute time)
          - y: y coords of mann box
          - z: z coords of mann box
        """
        nt,ny,nz = u.shape
        dt       = dx/U
        t        = np.arange(0, dt*(nt-0.5), dt)
        nt       = len(t)
        if y[0]>y[-1]:
            raise Exception('y is assumed to go from - to +')

        self['u']=np.zeros((3, nt, ny, nz))
        self['u'][0,:,:,:] = u 
        self['u'][1,:,:,:] = v
        self['u'][2,:,:,:] = w
        if addU is not None:
            self['u'][0,:,:,:] += addU
        self['t']  = t
        self['y']  = y
        self['z']  = z
        self['dt'] = dt
        # TODO
        self['ID'] = 7 # ...
        self['info'] = 'Converted from MannBox fields {:s}.'.format(time.strftime('%d-%b-%Y at %H:%M:%S', time.localtime()))
#         self['zTwr'] = np.array([])
#         self['uTwr'] = np.array([])
        self['zRef'] = None
        self['uRef'] = None
        self['zRef'], self['uRef'], bHub = self.hubValues()

    def toMannBox(self, base=None, removeUConstant=None, removeAllUMean=False):
        """ 
        removeUConstant: float,  will be removed from all values of the U box
        removeAllUMean: If true, the time-average of each y-z points will be substracted
        """
        try:
            from weio.mannbox_file import MannBoxFile
        except:
            try:
                from .mannbox_file import MannBoxFile
            except:
                from mannbox_file import MannBoxFile
        # filename
        if base is None:
            base = os.path.splitext(self.filename)[0]
        base = base+'_{}x{}x{}'.format(*self['u'].shape[1:])

        mn = MannBoxFile()
        mn.fromTurbSim(self['u'], 0, removeConstant=removeUConstant, removeAllMean=removeAllUMean)
        mn.write(base+'.u')

        mn.fromTurbSim(self['u'], 1)
        mn.write(base+'.v')

        mn.fromTurbSim(self['u'], 2)
        mn.write(base+'.w')

    # --- Useful IO
    def writeInfo(ts, filename):
        """ Write info to txt """
        infofile = filename
        with open(filename,'w') as f:
            f.write(str(ts))
            zMid =(ts['z'][0]+ts['z'][-1])/2
            f.write('Middle height of box: {:.3f}\n'.format(zMid))
            y_fit, pfit, model, _ = ts.fitPowerLaw(z_ref=zMid, y_span='mid', U_guess=10, alpha_guess=0.1)
            f.write('Power law: alpha={:.5f}  -  u={:.5f}  at z={:.5f}\n'.format(pfit[1],pfit[0],zMid))
            f.write('Periodic: {}\n'.format(ts.checkPeriodic(sigmaTol=1.5, aTol=0.5)))



    def writeProbes(ts, probefile, yProbe, zProbe):
        """ Create a CSV file with wind speed data at given probe locations
        defined by the vectors yProbe and zProbe. All combinations of y and z are extracted.
        INPUTS:
         - probefile: filename of CSV file to be written
         - yProbe: array like of y locations
         - zProbe: array like of z locations
        """
        Columns=['Time_[s]']
        Data   = ts['t']
        for y in yProbe:
            for z in zProbe:
                iy = np.argmin(np.abs(ts['y']-y))
                iz = np.argmin(np.abs(ts['z']-z))
                lbl = '_y{:.0f}_z{:.0f}'.format(ts['y'][iy], ts['z'][iz])
                Columns+=['{}{}_[m/s]'.format(c,lbl) for c in['u','v','w']]
                DataSub = np.column_stack((ts['u'][0,:,iy,iz],ts['u'][1,:,iy,iz],ts['u'][2,:,iy,iz]))
                Data    = np.column_stack((Data, DataSub))
        np.savetxt(probefile, Data, header=','.join(Columns), delimiter=',')

    def fitPowerLaw(ts, z_ref=None, y_span='full', U_guess=10, alpha_guess=0.1):
        """ 
        Fit power law to vertical profile
        INPUTS:
         - z_ref: reference height used to define the "U_ref"
         - y_span: if 'full', average the vertical profile across all y-values
                   if 'mid', average the vertical profile at the middle y value
        """
        if z_ref is None:
            # use mid height for z_ref
            z_ref =(ts['z'][0]+ts['z'][-1])/2
        # Average time series
        z, u, _ = ts.vertProfile(y_span=y_span)
        u = u[0,:]
        u_fit, pfit, model =  fit_powerlaw_u_alpha(z, u, z_ref=z_ref, p0=(U_guess, alpha_guess))
        return u_fit, pfit, model, z_ref



def fit_powerlaw_u_alpha(x, y, z_ref=100, p0=(10,0.1)):
    """ 
    p[0] : u_ref
    p[1] : alpha
    """
    import scipy.optimize as so
    pfit, _ = so.curve_fit(lambda x, *p : p[0] * (x / z_ref) ** p[1], x, y, p0=p0)
    y_fit = pfit[0] * (x / z_ref) ** pfit[1]
    coeffs_dict={'u_ref':pfit[0],'alpha':pfit[1]}
    formula = '{u_ref} * (z / {z_ref}) ** {alpha}'
    fitted_fun = lambda xx: pfit[0] * (xx / z_ref) ** pfit[1]
    return y_fit, pfit, {'coeffs':coeffs_dict,'formula':formula,'fitted_function':fitted_fun}

if __name__=='__main__':
    ts = TurbSimFile('../_tests/TurbSim.bts')